import subprocess
import sys
from pathlib import Path
from shutil import rmtree
from typing import Optional

import click
import yaml
from loguru import logger

from transpire.internal import render
from transpire.internal.config import ClusterConfig, get_config


@click.group()
def commands(**_) -> None:
    """tools related to Kubernetes objects"""
    pass


@commands.command()
@click.argument("out_path", envvar="TRANSPIRE_OBJECT_OUTPUT", type=click.Path())
def build(out_path, **_) -> None:
    """build objects, write them to a folder"""
    config = ClusterConfig.from_cwd()
    modules = [
        c.load_module_w_context(n, context=config) for n, c in config.modules.items()
    ]

    out_path = Path(out_path)
    out_path.mkdir(exist_ok=True, parents=True)

    for module in modules:
        logger.info(f"Building {module.name}")
        render.write_manifests(config, module.objects, module.name, out_path)

    logger.info("Writing bases")
    basedir = Path(out_path) / "base"
    if basedir.exists():
        rmtree(basedir)
    for module in modules:
        render.write_base(basedir, module)


@commands.command("print")
@click.argument("app_name", required=False)
def list_manifests(app_name: Optional[str] = None, **_) -> None:
    """build objects, print them to stdout"""
    module = get_config(app_name)
    yaml.safe_dump_all(module.objects, sys.stdout)


@commands.command()
@click.argument("app_name", required=True)
def apply(app_name: str, **_) -> None:
    """build objects, apply them to current kubernetes context"""
    # TODO: We should probably use the Kubernetes API here instead of shelling out.
    # That said, we do want this to behave exactly like kubectl apply, and what better
    # way to do that than to actually use kubectl apply? So maybe we don't want that.
    module = get_config(app_name)
    subprocess.run(
        ["kubectl", "create", "ns", module.namespace],
        check=False,
    )
    subprocess.run(
        ["kubectl", "apply", "-n", module.namespace, "-f", "-"],
        input=yaml.safe_dump_all(module.objects).encode("utf_8"),
    )



import importlib
import importlib.util
import os
import re
import shutil
import tomllib
from abc import ABC, abstractmethod
from functools import cache
from pathlib import Path
from subprocess import CalledProcessError, check_output
from types import ModuleType
from typing import Literal, Optional

from pydantic import AnyUrl, BaseModel, Field, HttpUrl

from transpire.internal.secrets import SecretsProvider
from transpire.internal.secrets.vault import HashicorpVaultConfig, VaultSecret
from transpire.types import Module


def first_env(*args: str, default: str | None = None) -> str:
    """
    try all environment variables in order, returning the first one that's set

    >>> import os
    >>> os.environ["VAR_THAT_EXISTS"] = "this exists!"
    >>> first_env("THIS_DOESNT_EXIST", "THIS_DOESNT_EITHER", "VAR_THAT_EXISTS")
    "this exists!"
    >>> first_env("THIS_DOESNT_EXIST", default="foo")
    "foo"
    """

    if len(args) == 0:
        if default is None:
            raise KeyError(
                "Unable to pull from environment, and no default was provided."
            )
        return default
    return os.environ.get(args[0], first_env(*args[1:], default=default))


class CLIConfig(BaseModel):
    """Configuration information for the transpire CLI tool."""

    git_path: Path = Field(description="Path to the git executable", default="git")
    cache_dir: Path = Field(description="The directory where cached files are stored")
    config_dir: Path = Field(
        description="The directory where transpire should write its persistent config files"
    )

    @classmethod
    @cache
    def from_env(cls) -> "CLIConfig":
        """pull configuration from environment variables, falling back to defaults as neccesary"""
        # TRANSPIRE_CACHE_DIR > XDG_CACHE_HOME > ~/.cache/
        cache_dir = (
            Path(first_env("TRANSPIRE_CACHE_DIR", "XDG_CACHE_HOME", default="~/.cache"))
            / "transpire"
        )

        # TRANSPIRE_CONFIG_DIR > XDG_CONFIG_HOME > ~/.config/
        config_dir = (
            Path(
                first_env(
                    "TRANSPIRE_CONFIG_DIR", "XDG_CONFIG_HOME", default="~/.config"
                )
            )
            / "transpire"
        )
        return cls(cache_dir=cache_dir.expanduser(), config_dir=config_dir.expanduser())


def load_py_module_from_file(
    py_mod_name: str, path: Path, expected_app_name: str | None = None
) -> ModuleType:
    spec = importlib.util.spec_from_file_location(py_mod_name, path)
    if spec is None or spec.loader is None:
        raise ValueError(f"No Python module was found at {path}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    try:
        real_app_name = module.name
    except AttributeError:
        raise ValueError(
            f"Python module at {path} does not appear to be a transpire module - missing `name'"
        )
    if expected_app_name is not None and real_app_name != expected_app_name:
        raise ValueError(
            f"Python module at {path} has wrong name: expected {expected_app_name}, got {real_app_name}"
        )

    return module


class ModuleConfig(ABC):
    @abstractmethod
    def load_module(self, name: str | None) -> Module:
        ...

    def load_module_w_context(self, name: str | None, context) -> Module:
        module = self.load_module(name)
        module.glob_context = context
        return module


class LocalModuleConfig(ModuleConfig, BaseModel):
    path: Path = Field(
        description="The path to the transpire config file within the module"
    )

    def load_module(self, name: str | None) -> Module:
        # TODO: do something about the implicit assumption that cwd == root of cluster repo
        # TODO: handle escaping file stem, make ".transpire.py" sane
        return Module(
            load_py_module_from_file(
                re.sub("[^A-Za-z0-9_]", "_", self.path.stem),
                self.path,
                name,
            )
        )

    @classmethod
    @cache
    def from_local(cls, path: Path | None = None) -> "LocalModuleConfig":
        raise NotImplementedError()


class GitModuleConfig(ModuleConfig, BaseModel):
    git: AnyUrl = Field(
        description="The URL of the remote git repository the module resides in"
    )
    branch: str | None = Field(description="The branch to deploy from")
    dir: Path = Field(
        description="The root path containing the module", default=Path(".")
    )

    @property
    def resolved_dir(self) -> Path:
        if self.dir.is_absolute():
            return self.dir.relative_to("/")
        return self.dir

    @property
    def clean_git_url(self):
        return self.git.removesuffix(".git") + ".git"

    def get_cached_repo(self, *, commit: str | None = None) -> tuple[Path, str]:
        config = CLIConfig.from_env()
        cache_root = config.cache_dir / "remote_modules"
        cache_dir = cache_root / re.sub("[^A-Za-z0-9]", "_", self.git)
        cache_root.mkdir(exist_ok=True, parents=True)

        def call_cached_git(*args):
            return check_output([config.git_path, *args], cwd=cache_dir)

        if commit is None:
            fetch_args = [self.branch or "HEAD", "--depth", "1"]
            branch_args = [] if self.branch is None else ["--branch", self.branch]
            clone_args = ["--depth", "1", "--single-branch", *branch_args]
        else:
            # FIXME lol --unshallow is not idempotent
            fetch_args = ["--depth=10000000"]
            clone_args = []

        if cache_dir.exists():
            try:
                call_cached_git("fetch", self.git, *fetch_args)
                call_cached_git("checkout", "--detach")
                call_cached_git("reset", "--hard", commit or "FETCH_HEAD")
                call_cached_git("clean", "-dfx")
            except CalledProcessError:
                shutil.rmtree(cache_dir)
            else:
                return cache_dir, call_cached_git("rev-parse", "HEAD").decode().strip()

        cache_dir.mkdir(exist_ok=True, parents=True)
        check_output([config.git_path, "clone", *clone_args, self.git, cache_dir])

        call_cached_git("checkout", "--detach")
        if commit is None:
            commit = call_cached_git("rev-parse", "HEAD").decode().strip()
        else:
            call_cached_git("reset", "--hard", commit)

        return cache_dir, commit

    def load_module(self, name: str | None, *, commit: str | None = None) -> Module:
        cache_dir, commit = self.get_cached_repo(commit=commit)
        py_module = load_py_module_from_file(
            "_transpire", cache_dir / self.resolved_dir / ".transpire.py", name
        )
        module = Module(py_module)
        module.revision = commit
        return module


class CIConfig(BaseModel):
    """CI Configuration"""

    namespace: str = Field(description="kubernetes namespace", default="transpire")
    webhook_url: HttpUrl = Field(description="github webhook url")


class ClusterDefaults(BaseModel):
    ingressClass: str | None
    certManagerIssuer: str | None


class ClusterConfig(BaseModel):
    """Cluster configuration"""

    class SecretsConfig(BaseModel):
        provider: Literal["vault"] = Field(description="secrets provider to use")
        vault: Optional[HashicorpVaultConfig] = Field(
            description="configuration for bitnami sealed secrets"
        )

    apiVersion: Literal["v1"]
    secrets: SecretsConfig = Field(description="configuration for the secrets provider")
    modules: dict[
        str,
        LocalModuleConfig | GitModuleConfig,
    ] = Field(description="list of modules to load")

    ci: CIConfig
    defaults: ClusterDefaults

    @classmethod
    @cache
    def from_cwd(cls, cwd: Path = Path.cwd()) -> "ClusterConfig":
        cluster_toml = cwd / "cluster.toml"
        if cluster_toml.exists():
            return cls.parse_obj(tomllib.loads(cluster_toml.read_text()))
        if cwd.is_mount() or (cwd / ".git").exists():
            raise FileNotFoundError(
                "cluster.toml not found up to current git or fs boundary"
            )
        return cls.from_cwd(cwd.parent)


def get_config(module_name: str | None = None, cwd: Path = Path.cwd()) -> Module:
    if module_name:
        cluster_config = ClusterConfig.from_cwd(cwd)
        module = cluster_config.modules.get(module_name)
        if module is not None and isinstance(
            module, (LocalModuleConfig, GitModuleConfig)
        ):
            return module.load_module_w_context(module_name, context=cluster_config)
        cluster_toml = cwd / "cluster.toml"
        raise ValueError(
            f"Python module at {cluster_toml} is missing module {module_name}"
        )

    transpire_py = cwd / ".transpire.py"
    if transpire_py.exists():
        return Module(load_py_module_from_file("_transpire", transpire_py, None))
    if cwd.is_mount() or (cwd / ".git").exists():
        raise FileNotFoundError(
            ".transpire.py not found up to current git or fs boundary"
        )
    return get_config(module_name, cwd.parent)


def provider_from_context(
    namespace: str, dev: bool = False, config: ClusterConfig | None = None
) -> SecretsProvider:
    if not config:
        config = ClusterConfig.from_cwd()
    if config.secrets.provider == "vault":
        if not config.secrets.vault:
            raise ValueError("Options for Vault not provided.")
        return VaultSecret(config.secrets.vault, namespace, dev)
    raise ValueError("No valid secrets provider found.")



from contextvars import ContextVar

# TODO: fixme
from typing import TYPE_CHECKING, ParamSpec, TypeVar

if TYPE_CHECKING:
    from transpire.internal.config import ClusterConfig
    from transpire.types import Module


_app_context: ContextVar["Module"] = ContextVar("current_app")
_current_global: ContextVar["ClusterConfig"] = ContextVar("current_global")

T = TypeVar("T")
P = ParamSpec("P")


def set_app_context(info: "Module") -> None:
    _app_context.set(info)


def set_global_context(context: "ClusterConfig") -> None:
    _current_global.set(context)


def get_app_context() -> "Module":
    return _app_context.get()


def get_global_context() -> "ClusterConfig":
    return _current_global.get()



import click

from . import bootstrap, dev, image, obj, secrets


@click.group()
def cli() -> None:
    pass


cli.add_command(bootstrap.commands, "bootstrap")
cli.add_command(dev.commands, "dev")
cli.add_command(obj.commands, "object")
cli.add_command(image.commands, "image")
cli.add_command(secrets.commands, "secret")



from typing import Any, Callable, Iterable, Optional, cast

__all__ = ["delve", "shelve", "edit_manifests", "make_edit_manifest"]

RESOURCE_APIS = {
    "Deployment": "apps/v1",
    "Secret": "v1",
    "Service": "v1",
    "ConfigMap": "v1",
}


def delve(obj: dict, path: Iterable[str]) -> Optional[Any]:
    """
    Get the element of the nested dict at the path provided, returning None if
    any keys along the way do not exist
    """
    curr = obj
    for key in path:
        try:
            curr = curr[key]
        except KeyError:
            return None
    return curr


def shelve(
    obj: dict, path: Iterable[str], val: Any, *, create_parents: bool = False
) -> dict:
    """
    Set the element of the nested dict at the path provided, returning the
    (in-place modified) object, throwing KeyError if any keys along the way do
    not exist
    """
    curr: Optional[dict] = obj
    prev: Optional[dict] = None
    prev_key = None
    for key in path:
        if curr is None:
            if create_parents:
                curr = dict()
                assert prev is not None
                prev[prev_key] = curr
            else:
                raise KeyError(prev_key)
        prev, prev_key = curr, key
        try:
            curr = curr[key]
        except KeyError:
            curr = None
    if prev is None:
        return val
    prev[prev_key] = val
    return obj


def edit_manifests(
    edits: dict[
        tuple[str, str] | tuple[tuple[str, str], str], Callable[[dict], dict | None]
    ],
    manifests: Iterable[dict],
) -> list[dict]:
    resolved_edits: dict[tuple[str | None, str, str], Callable[[dict], dict | None]] = {
        (
            k[0][0]
            if isinstance(k[0], tuple)
            else RESOURCE_APIS.get(k[0], cast(str, None)),
            k[0] if isinstance(k[0], str) else k[0][1],
            k[1],
        ): v
        for k, v in edits.items()
    }
    unseen = set(resolved_edits.keys())

    def mapper(m: dict) -> dict | None:
        key1: tuple[str | None, str, str] = (
            m["apiVersion"],
            m["kind"],
            m["metadata"]["name"],
        )
        key2 = (None, *key1[1:])
        func = resolved_edits.get(key1, resolved_edits.get(key2, None))
        if func is not None:
            try:
                unseen.remove(key1)
            except KeyError:
                unseen.remove(key2)
            return func(m)
        return m

    result = list(o for o in map(mapper, manifests) if o is not None)
    if unseen:
        raise RuntimeError(f"Some edits were not applied: {repr(unseen)}")
    return result


def make_edit_manifest(
    edits: dict[Iterable[str], Any], *, create_parents: bool = False
) -> Callable[[dict], dict]:
    def edit(m):
        for path, val in edits.items():
            shelve(m, path, val, create_parents=create_parents)
        return m

    return edit



import click

from transpire.internal import config


@click.group()
def commands(**_):
    """tools to aid development with transpire"""
    pass


@commands.command()
def schema(**_) -> None:
    """print the json schema for a transpire parent"""
    print(config.ClusterConfig.schema_json())



import os
from datetime import datetime, timezone

import click

from transpire.internal.config import ClusterConfig, GitModuleConfig
from transpire.types import Image, Module

REGISTRY = "harbor.ocf.berkeley.edu"


def image_metadata(config: GitModuleConfig, module: Module, image: Image):
    base_tag = f"{REGISTRY}/ocf/{module.name}/{image.name}"
    return {
        "tags": [
            f"{base_tag}:latest",
            f"{base_tag}:{module.revision}",
        ],
        "labels": {
            "org.opencontainers.image.url": config.clean_git_url,
            "org.opencontainers.image.source": config.clean_git_url,
            "org.opencontainers.image.created": datetime.now(timezone.utc).isoformat(),
            "org.opencontainers.image.revision": module.revision,
        },
    }


@click.group()
def commands(**_) -> None:
    """tools related to images"""
    pass


@commands.command()
@click.argument("module_name", required=True)
@click.option("-o", "--output", required=True, type=click.Choice(["gha"]))
@click.option("--commit")
def build(module_name: str, output: str, commit: str | None) -> None:
    """build images"""

    if output == "gha":
        config = ClusterConfig.from_cwd()
        module_config = config.modules[module_name]

        if not isinstance(module_config, GitModuleConfig):
            click.echo("Building images is only supported for git modules", err=True)
            return

        module = module_config.load_module(module_name, commit=commit)

        git_url = f"{module_config.clean_git_url}#"
        if module_config.branch is not None:
            git_url += module_config.branch

        images = [
            {
                "name": x.name,
                "context": f"{git_url}:{x.resolved_path}",
                **image_metadata(module_config, module, x),
            }
            for x in module.images
        ]

        print(os.environ["GITHUB_OUTPUT"])
        print(f"image_matrix={images}")

        with open(os.environ["GITHUB_OUTPUT"], "a") as f:
            print(f"image_matrix={images}", file=f)



from .configmap import ConfigMap
from .deployment import Deployment
from .ingress import Ingress
from .secret import Secret
from .service import Service

__all__ = ["Deployment", "Ingress", "Service", "Secret", "ConfigMap"]



from typing import Union

from kubernetes import client

from transpire.resources.base import Resource


class Service(Resource[client.V1Service]):
    def __init__(
        self,
        name: str,
        selector: dict[str, str],
        port_on_pod: Union[int, str],
        port_on_svc: Union[int, str],
    ):
        self.obj = client.V1Service(
            api_version="v1",
            kind="Service",
            metadata=client.V1ObjectMeta(name=name),
            spec=client.V1ServiceSpec(
                selector=selector,
                ports=[client.V1ServicePort(port=port_on_svc, target_port=port_on_pod)],
            ),
        )
        super().__init__()



from collections.abc import Callable, Iterable

from transpire.internal.surgery import delve
from transpire.internal.surgery import edit_manifests as _edit_manifests
from transpire.internal.surgery import make_edit_manifest, shelve
from transpire.manifestlike import ManifestLike, manifests_to_dict


def edit_manifests(
    edits: dict[
        tuple[str, str] | tuple[tuple[str, str], str], Callable[[dict], dict | None]
    ],
    manifests: ManifestLike | Iterable[ManifestLike | None],
) -> list[dict]:
    return _edit_manifests(edits, manifests_to_dict(manifests))


__all__ = ["delve", "shelve", "edit_manifests", "make_edit_manifest"]



from typing import Callable, Generic, Self, TypeVar

from transpire.manifestlike import ManifestLike, manifest_to_dict

T = TypeVar("T", bound=ManifestLike)


class Resource(Generic[T]):
    obj: T
    patches: list

    def __init__(self):
        self.patches = []

    def patch(self, *fns: Callable[[dict], dict]) -> Self:
        self.patches.extend(fns)
        return self

    def build(self) -> dict:
        out = manifest_to_dict(self.obj)
        for patch in self.patches:
            out = patch(out)
        return out



import shutil
import tempfile
from subprocess import PIPE, run
from typing import Any

import yaml

from transpire.internal.config import CLIConfig
from transpire.internal.context import get_app_context

__all__ = ["build_chart_from_versions", "build_chart"]


def assert_helm() -> None:
    """ensure helm binary is available"""

    if shutil.which("helm") is None:
        raise RuntimeError("`helm` must be installed and in your $PATH")


def exec_helm(args: list[str], check: bool = True) -> tuple[bytes, bytes]:
    """executes a helm command and returns (stdout, stderr)"""

    config = CLIConfig.from_env()
    process = run(
        [
            "helm",
            "--registry-config",
            str(config.cache_dir / "helm" / "registry.json"),
            "--repository-cache",
            str(config.cache_dir / "helm" / "repository"),
            "--repository-config",
            str(config.cache_dir / "helm" / "repositories.yaml"),
            *args,
        ],
        check=False,
        stdout=PIPE,
        stderr=PIPE,
    )

    if check and process.returncode != 0:
        raise ValueError(process.stderr)

    return (process.stdout, process.stderr)


def add_repo(name: str, url: str) -> None:
    """add a repository to transpire's Helm repository list"""

    assert_helm()
    _, stderr = exec_helm(["repo", "add", name, url], check=True)
    if stderr:
        print(stderr.decode("utf-8"))


def update_repo(name: str) -> None:
    """update a repository in transpire's Helm cache"""

    assert_helm()
    exec_helm(["repo", "update", name], check=False)


def build_chart(
    repo_url: str,
    chart_name: str,
    name: str,
    version: str,
    values: dict | None = None,
    capabilities: list[str] | None = None,
) -> list[dict]:
    """build a helm chart and return a list of manifests"""

    # TODO: avoid needing to setting capabilities for "normal" things
    # - maybe have a config file at cluster level?

    add_repo(name, repo_url)
    update_repo(name)

    with tempfile.NamedTemporaryFile(suffix=".yml") as values_file:
        values_file.write(yaml.dump(values).encode("utf-8"))
        values_file.flush()

        capabilities_flag = []
        if capabilities is not None and len(capabilities) > 0:
            capabilities_flag = ["--api-versions", ", ".join(capabilities)]

        # TODO: Capture `stderr` output and make available to tracing.
        stdout, _ = exec_helm(
            [
                "template",
                "-n",
                get_app_context().namespace,
                "--values",
                values_file.name,
                "--include-crds",
                "--version",
                version,
                "--name-template",
                name,
                *capabilities_flag,
                f"{name}/{chart_name}",
            ],
            check=True,
        )

    # save our souls
    # <https://github.com/prometheus-community/helm-charts/pull/2238>
    # <https://github.com/prometheus-operator/prometheus-operator/pull/4897>
    # <https://github.com/yaml/pyyaml/issues/89>
    # <https://github.com/allenporter/k8s-gitops/commit/304c64c57926d2747328c0803c246be7dd827fdd>
    yaml.constructor.SafeConstructor.add_constructor(
        "tag:yaml.org,2002:value", yaml.constructor.SafeConstructor.construct_yaml_str  # type: ignore
    )

    return list(yaml.safe_load_all(stdout))


def build_chart_from_versions(
    name: str,
    versions: dict[str, Any],
    values: dict = {},
) -> list[dict]:
    """thin wrapper around build_chart that builds based off a versions dict"""

    return build_chart(
        repo_url=versions[name]["helm"],
        chart_name=versions[name].get("chart", name),
        name=name,
        version=versions[name]["version"],
        values=values,
    )



import textwrap
from pathlib import Path

import click
from loguru import logger


@click.group()
def commands(**_):
    """tools related to bootstrapping (new cluster, new repository, etc)"""
    pass


@commands.command()
def kubernetes(**_):
    """bootstrap kubernetes to pull from your cluster repository"""
    return NotImplemented


@commands.command()
@click.argument("path", type=click.Path(exists=True), default=".", required=False)
@click.option("-f", "--force", is_flag=True)
def repository(path, force, **_) -> None:
    """initializes current working directory to be a transpire app repository"""
    path = Path(path)
    transpire_py = path / ".transpire.py"

    if (not force) and transpire_py.exists():
        logger.error(
            "Looks like this repository is already initialized. To reinitialize, rerun with --force."
        )
        return

    transpire_py.write_text(
        textwrap.dedent(
            """
            from transpire.resources import Deployment

            \"\"\"
            This is the name of the current transpire module, which is used for:
            1. The name of the ArgoCD `Application` CRD which owns these resources.
            2. The default namespace that resources should be deployed to (i.e. if `namespace` isn't set on a resource, it won't be overridden).
            \"\"\"
            name = "echoserver"


            def objects():
                yield Deployment.simple(name=name, image="k8s.gcr.io/echoserver", ports=["8080"])
            """
        )
    )
    logger.info(
        f"Successfully initialized transpire app repository in {path.resolve().name}."
    )



from typing import List, Self, Union

from kubernetes import client

from transpire.resources.base import Resource


class Deployment(Resource[client.V1Deployment]):
    def __init__(
        self,
        name: str,
        image: str,
        ports: List[Union[str, int]],
    ):
        self.obj = client.V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=client.V1ObjectMeta(name=name),
            spec=client.V1DeploymentSpec(
                replicas=1,
                selector=client.V1LabelSelector(
                    match_labels={"transpire.ocf.io/deployment": name}
                ),
                template=client.V1PodTemplateSpec(
                    metadata=client.V1ObjectMeta(
                        labels={"transpire.ocf.io/deployment": name}
                    ),
                    spec=client.V1PodSpec(
                        containers=[
                            client.V1Container(
                                name="main",
                                image=image,
                                image_pull_policy="IfNotPresent",
                                ports=[
                                    client.V1ContainerPort(container_port=x)
                                    for x in ports
                                ],
                            )
                        ]
                    ),
                ),
            ),
        )
        super().__init__()

    def _init_env(self, container_id: int) -> client.V1Container:
        container = self.obj.spec.template.spec.containers[container_id]

        if container.env_from is None:
            container.env_from = []
        if container.env is None:
            container.env = []

        return container

    def with_configmap_env(
        self, name: str, *, mapping: dict[str, str] | None = None, container_id: int = 0
    ) -> Self:
        container = self._init_env(container_id)
        if mapping is None:
            container.env_from.append(
                client.V1EnvFromSource(
                    config_map_ref=client.V1ConfigMapEnvSource(
                        name=name,
                    )
                )
            )
        else:
            container.env.extend(
                client.V1EnvVar(
                    name=envvar_name,
                    value_from=client.V1EnvVarSource(
                        config_map_key_ref=client.V1ConfigMapKeySelector(
                            name=name,
                            key=cm_key,
                        )
                    ),
                )
                for envvar_name, cm_key in mapping.items()
            )
        return self

    def with_secrets_env(
        self, name: str, *, mapping: dict[str, str] | None = None, container_id: int = 0
    ) -> Self:
        container = self._init_env(container_id)
        if mapping is None:
            container.env_from.append(
                client.V1EnvFromSource(
                    secret_ref=client.V1SecretEnvSource(
                        name=name,
                    )
                )
            )
        else:
            container.env.extend(
                client.V1EnvVar(
                    name=envvar_name,
                    value_from=client.V1EnvVarSource(
                        secret_key_ref=client.V1SecretKeySelector(
                            name=name,
                            key=secret_key,
                        )
                    ),
                )
                for envvar_name, secret_key in mapping.items()
            )
        return self

    def get_container(
        self, name: str | None = None, *, remove: bool = False
    ) -> client.V1Container:
        container_list = self.obj.spec.template.spec.containers

        if name is None:
            if len(container_list) != 1:
                raise ValueError("If multiple containers, must pass name.")
            return container_list[0]

        for i, container in enumerate(container_list):
            if container.name == name:
                if remove:
                    return container_list.pop(i)
                return container

        raise ValueError(f"No such container: {name}")

    def add_container(self, container: client.V1Container) -> int:
        container_list = self.obj.spec.template.spec.containers

        names = set(c.name for c in container_list)
        if container.name in names:
            raise ValueError(f"Can't use name {container.name}, already in use.")

        container_list.append(container)
        return len(container_list) - 1



from transpire.internal.config import ClusterConfig, provider_from_context
from transpire.manifestlike import ManifestLike


class ManifestError(Exception):
    def __init__(self, message: str, *, suggestion: ManifestLike | None) -> None:
        self.message = message
        self.suggestion = suggestion


def postprocess(
    config: ClusterConfig, obj: dict, appname: str, dev: bool = False
) -> dict:
    """run all postprocessing steps (right now just secret processing)"""

    provider = provider_from_context(appname, dev=dev, config=config)
    if obj["apiVersion"] == "v1" and obj["kind"] == "Secret":
        return provider.convert_secret(obj)

    return obj



from kubernetes import client

from transpire.resources.base import Resource


class ConfigMap(Resource[client.V1ConfigMap]):
    def __init__(
        self,
        name: str,
        *,
        data: dict[str, str],
    ):
        self.obj = client.V1ConfigMap(
            api_version="v1",
            kind="ConfigMap",
            metadata=client.V1ObjectMeta(
                name=name,
            ),
            data=data,
        )
        super().__init__()



from abc import ABC, abstractmethod

from transpire.manifestlike import ManifestLike


class SecretsProvider(ABC):
    @abstractmethod
    def convert_secret(self, secret: ManifestLike) -> dict:
        ...

    @abstractmethod
    def push_secret(self, secret: ManifestLike) -> None:
        ...



import base64
import os

import hvac  # type: ignore
import hvac.exceptions  # type: ignore
from pydantic import BaseModel

from transpire.internal.secrets import SecretsProvider
from transpire.manifestlike import ManifestLike, manifest_to_dict


class HashicorpVaultConfig(BaseModel):
    kvstore: str


def fix_base64(pairs: dict) -> dict:
    """Runs base64^-1(v) on every value in a dict."""
    return {k: base64.b64decode(v).decode() for k, v in pairs.items()}


def extract_secret(secret: dict) -> dict:
    """Takes a Kubernetes secret object, returns analogous Vault path and the k:v secrets."""
    # TODO: Handle generateName? Though who would put generateName on a secret? Does that even work???
    return {
        **fix_base64(safe_get(secret, "data")),
        **safe_get(secret, "stringData"),
    }


def safe_get(manifest: dict, key: str) -> dict:
    """
    A version of .get(key, {}).items() that correctly handles
    the case where manifest[key] is None, and not that it doesn't
    exist.
    """
    out = manifest.get(key, {})
    return out if out else {}


class VaultSecret(SecretsProvider):
    def __init__(
        self, config: HashicorpVaultConfig, ns: str, dev: bool = False
    ) -> None:
        self.kvstore = config.kvstore
        self.dev = dev
        self.ns = ns
        # TODO: Make this toggle-offable.
        self.push = dev

    def push_secret(self, secret: ManifestLike) -> None:
        """v1/Secret -> Hashicorp Vault"""
        secret = manifest_to_dict(secret)
        path = f"{self.ns}/{secret['metadata']['name']}"
        client = hvac.Client("https://vault.ocf.berkeley.edu")
        client.token = os.getenv("VAULT_TOKEN")
        if not client.token:
            print(
                "I couldn't find your VAULT_TOKEN env variable, so I can't push secrets to Vault.",
            )
        assert client.is_authenticated()
        client.secrets.kv.v2.configure(mount_point=self.kvstore)

        pairs = extract_secret(secret)
        try:
            client.secrets.kv.v2.create_or_update_secret(
                path=path,
                secret=pairs,
                cas=0,
                mount_point=self.kvstore,
            )
        except hvac.exceptions.InvalidRequest as e:
            print(f"{e} -- Probably '{path}' already created, not re-creating")

    def convert_secret(self, secret: ManifestLike) -> dict:
        """v1/Secret -> ricoberger.de/v1alpha1/SealedSecret"""
        secret = manifest_to_dict(secret)
        path = f"{self.ns}/{secret['metadata']['name']}"

        if self.dev:
            return secret

        keys = [
            *[k for k, _ in safe_get(secret, "data").items()],
            *[k for k, _ in safe_get(secret, "stringData").items()],
        ]

        return {
            "apiVersion": "ricoberger.de/v1alpha1",
            "kind": "VaultSecret",
            "metadata": {"name": secret["metadata"]["name"]},
            "spec": {
                "keys": keys,
                "path": f"{self.kvstore}/{path}",
                "type": "Opaque",
            },
        }



from transpire.internal.validation import is_valid_dnsname


def make_app(
    app_name: str,
    app_namespace: str,
    repo_url: str = "https://github.com/ocf/cluster.git",
    repo_branch: str = "HEAD",
) -> dict:
    if not is_valid_dnsname(app_name):
        raise ValueError(f"Expected a valid DNS name, but got {app_name} instead.")

    return {
        "apiVersion": "argoproj.io/v1alpha1",
        "kind": "Application",
        "metadata": {"name": app_name, "namespace": "argocd"},
        "spec": {
            "project": "default",
            "destination": {
                "server": "https://kubernetes.default.svc",
                "namespace": app_namespace,
            },
            "source": {
                "repoURL": repo_url,
                "path": f"{app_name}",
                "targetRevision": repo_branch,
            },
            "syncPolicy": {
                "syncOptions": [
                    "CreateNamespace=true",
                ],
            },
        },
    }



import pathlib
import tomllib
from typing import Dict, cast


def get_file(caller, filename: str) -> pathlib.Path:
    """
    Given the __file__ parameter of the calling file and
    the filename of an adjacent file, return the path to
    that file.
    """
    return pathlib.Path(caller).parent.resolve() / filename


def get_versions(caller) -> Dict[str, Dict[str, str]]:
    """
    Given the __file__ parameter of the calling file, return
    the contents of versions.toml cast to a dict.
    """
    file = get_file(caller, "versions.toml")

    return cast(
        Dict[str, Dict[str, str]],
        tomllib.loads(file.read_text()),
    )



from pathlib import Path
from shutil import rmtree
from typing import Iterable

import yaml
from loguru import logger

from transpire.internal import argocd
from transpire.internal.config import ClusterConfig
from transpire.internal.postprocessor import ManifestError, postprocess
from transpire.types import Module


def write_manifests(
    config: ClusterConfig, objects: Iterable[dict], appname: str, manifest_dir: Path
) -> None:
    """Write objects to manifest_dir as YAML files."""
    appdir = manifest_dir / appname
    if appdir.exists():
        rmtree(appdir)
    appdir.mkdir(exist_ok=True)

    failed = False
    processed_objs = {}

    for obj in objects:
        try:
            obj = postprocess(config, obj, appname, dev=False)
        except ManifestError as err:
            name = obj["metadata"].get("name", obj["metadata"].get("generateName"))
            logger.exception(f"Error processing object: {name}")
            if err.suggestion:
                logger.info("Suggested replacement:")
                logger.info(err.suggestion)
            failed = True
            continue

        name = obj["metadata"].get("name", obj["metadata"].get("generateName", None))
        kind = obj["kind"]
        namespace = obj["metadata"].get("namespace", appname)
        if (
            obj["kind"] == "SyncedSecret"
            and (appdir / f"{name}_{kind}_{namespace}.yaml").exists()
        ):
            continue
        processed_objs[f"{name}_{kind}_{namespace}.yaml"] = obj

    if failed:
        logger.error("Exceptions encountered, manifests will not be written.")
    else:
        for fname, obj in processed_objs.items():
            with open(appdir / fname, "w") as f:
                yaml.safe_dump(obj, f)


def write_base(basedir: Path, module: Module):
    basedir.mkdir(exist_ok=True)
    obj = argocd.make_app(module.name, module.namespace)
    argo_namespace = obj["metadata"].get("namespace")
    if not argo_namespace:
        raise ValueError("Argo Application has unset namespace.")
    with open(basedir / f"{module.name}_Application_{argo_namespace}.yaml", "w") as f:
        yaml.safe_dump(obj, f)



from __future__ import annotations

from typing import Optional, Union

from kubernetes import client

from transpire.internal.context import get_global_context
from transpire.resources.base import Resource

from .service import Service


class Ingress(Resource[client.V1Ingress]):
    @classmethod
    def from_svc(cls, svc: Service, host: str, path_prefix: str = "/"):
        built = svc.build()
        # TODO: Make sure there's only one port.
        return cls(
            host=host,
            service_name=built["metadata"]["name"],
            service_port=built["spec"]["ports"][0]["port"],
            path_prefix=path_prefix,
        )

    def __init__(
        self,
        *,
        host: str,
        service_name: str,
        service_port: Union[int, str],
        ingress_name: Optional[str] = None,
        path_prefix: str = "/",
    ):
        ctx = get_global_context()
        self.obj = client.V1Ingress(
            api_version="networking.k8s.io/v1",
            kind="Ingress",
            metadata=client.V1ObjectMeta(
                name=ingress_name if ingress_name else service_name,
                annotations={
                    "cert-manager.io/cluster-issuer": ctx.defaults.certManagerIssuer,
                    "ingress.kubernetes.io/force-ssl-redirect": "true",
                    "io.cilium/websocket": "enabled",
                    "kubernetes.io/tls-acme": "true",
                    "projectcontour.io/websocket-routes": path_prefix,
                },
            ),
            spec=client.V1IngressSpec(
                ingress_class_name=ctx.defaults.ingressClass,
                rules=[
                    client.V1IngressRule(
                        host=host,
                        http=client.V1HTTPIngressRuleValue(
                            paths=[
                                client.V1HTTPIngressPath(
                                    path=path_prefix,
                                    path_type="Prefix",
                                    backend=client.V1IngressBackend(
                                        service=client.V1IngressServiceBackend(
                                            port=client.V1ServiceBackendPort(
                                                number=service_port,
                                            ),
                                            name=service_name,
                                        )
                                    ),
                                )
                            ]
                        ),
                    )
                ],
                tls=[
                    client.V1IngressTLS(
                        hosts=[host],
                        secret_name=f"{service_name}-tls",
                    )
                ],
            ),
        )
        super().__init__()



from typing import Iterable, Protocol

from kubernetes import client


class OpenAPIObject(Protocol):
    """An object that can be converted to a dictionary. Useful for objects from github.com/kubernetes-client/python."""

    openapi_types: dict
    attribute_map: dict


# Something that is possibly a Kubernetes manifest. No validation is performed.
ManifestLike = dict | OpenAPIObject
_api_client = client.ApiClient()


def manifest_to_dict(obj: ManifestLike) -> dict:
    try:
        sanitized = _api_client.sanitize_for_serialization(obj)
        assert isinstance(sanitized, dict)
        return sanitized
    except AttributeError:
        pass
    raise TypeError(f"unsupported manifest type: {type(obj)}")


def manifests_to_dict(
    objs: ManifestLike | Iterable[ManifestLike | None],
) -> Iterable[dict]:
    objs_iter: Iterable[ManifestLike | None]
    if isinstance(objs, dict):
        objs_iter = [objs]
    else:
        if isinstance(objs, Iterable):
            objs_iter = objs
        else:
            objs_iter = [objs]

    return (manifest_to_dict(o) for o in objs_iter if o is not None)



from collections.abc import Iterable
from contextvars import Context
from functools import cached_property
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, List, TypeVar

from pydantic import BaseModel, Field

from transpire.internal import context
from transpire.manifestlike import manifests_to_dict

_T = TypeVar("_T")


class Version(BaseModel):
    """An adelie (github.com/nikhiljha/adelie) compatible version definition."""

    version: str
    github: str | None = Field(
        description="A GitHub repository URL, used for release version checking."
    )
    helm: str | None
    chart: str | None


class Image(BaseModel):
    """Describes how to build an OCI image."""

    name: str
    path: Path

    @property
    def resolved_path(self) -> Path:
        if self.path.is_absolute():
            return self.path.relative_to("/")
        return self.path


class Module:
    """Transpire modules contain information about how to build and deploy applications to Kubernetes."""

    pymodule: ModuleType
    revision: str | None

    def __init__(self, pymodule: ModuleType, context=None):
        self.pymodule = pymodule
        self.glob_context = context

    @property
    def name(self) -> str:
        name = self.pymodule.name
        assert name != "base"
        return name

    @cached_property
    def namespace(self) -> str:
        if hasattr(self.pymodule, "namespace"):
            return self.pymodule.namespace
        return self.pymodule.name

    def _enter_context(self) -> None:
        context.set_app_context(self)
        if self.glob_context is not None:
            context.set_global_context(self.glob_context)

    def _render_fn(
        self, function: str, finalizer: Callable[[Any], _T], default: _T
    ) -> _T:
        def _list() -> Any:
            self._enter_context()
            if hasattr(self.pymodule, function):
                return finalizer(getattr(self.pymodule, function)())
            return default

        return Context().run(_list)

    def _render_iter(self, function: str) -> list[Any]:
        def finalizer(gen: Any) -> list[Any]:
            if not isinstance(gen, Iterable):
                raise ValueError(f"function `{function}` must be iterable")
            return list(gen)

        return self._render_fn(function=function, finalizer=finalizer, default=[])

    @cached_property
    def images(self) -> List[Image]:
        return list(self._render_iter("images"))

    @cached_property
    def objects(self) -> List[dict]:
        return list(manifests_to_dict(self._render_iter("objects")))
